/*-
 * Copyright (c) 2005 Andriy Gapon <avg@icyb.net.ua>
 * All rights reserved.
 *
 * Portions of this code are derived from
 * udfclient.c by Reinoud Zandijk <reinoud@netbsd.org>
 *
 * This code is a FUSE adapter to, and compiled with, and linked with
 * the code created and copyrighted by Reinoud Zandijk <reinoud@netbsd.org>
 * that is distributed under the Clarified Artistic Licence,
 * please see that code for details.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: udfclientfs.c,v 1.34 2009/01/08 18:33:24 avg Exp $
 */

#include <sys/types.h>
#include <sys/limits.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <paths.h>
#include <libgen.h>
#include <limits.h>
#include <errno.h>

#define FUSE_USE_VERSION 27
#include <fuse.h>

#include "udf.h"
#include "udf_bswap.h"

static int udfclientfs_debug = 0;

/* XXX setattr is not exported through udfclient headers */
extern int	udf_setattr(struct udf_node *udf_node, struct stat *stat);
extern void	udf_dump_alive_sets(void);
extern int	udf_start_unix_thread(void);

static int	get_node(const struct fuse_file_info *fi, struct udf_node **node);
static int	set_node(struct fuse_file_info *fi, const struct udf_node *node);
static int	decompose_path(const char *path, char **ppath, char **cpath);
static int	lookup_pathname(struct udf_node **node, const char *path);
static void	fixup_stat(const struct udf_node *node, struct stat *stat);
static int	udfclientfs_getattr(const char *path, struct stat *stat);
static int	udfclientfs_chmod(const char *path, mode_t mode);
static int	udfclientfs_chown(const char *path, uid_t uid, gid_t gid);
static int	udfclientfs_utime(const char *path, struct utimbuf *times);
static int	udfclientfs_truncate(const char *path, off_t length);
static int	udfclientfs_mknod(const char *path, mode_t mode, dev_t unused);
static int	udfclientfs_mkdir(const char *path, mode_t mode);
static int	udfclientfs_unlink(const char *path);
static int	udfclientfs_rmdir(const char *path);
static int	udfclientfs_rename(const char *from, const char *to);
static int	udfclientfs_open(const char *path, struct fuse_file_info *fi);
static int	udfclientfs_release(const char *path, struct fuse_file_info *fi);
static int	udfclientfs_read(const char *path, char *rbuf, size_t size, off_t offset, struct fuse_file_info *fi);
static int	udfclientfs_write(const char *path, const char *wbuf, size_t size, off_t offset, struct fuse_file_info *fi);
static int	udfclientfs_opendir(const char *path, struct fuse_file_info *fi);
static int	udfclientfs_releasedir(const char *path, struct fuse_file_info *fi);
static int	udfclientfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi);
static int	udfclientfs_fsync(const char *unused1, int unused2, struct fuse_file_info *unused3);
static int	udfclientfs_statfs(const char *unsed, struct statvfs *buf);
static void *	udfclientfs_init();
static void	udfclientfs_destroy(void *unused);

static struct udf_mountpoint *mp = NULL;
static struct udf_node *root_node = NULL;
static struct udf_discinfo *disc = NULL;
static struct stat mp_stat;

static int
get_node(const struct fuse_file_info *fi, struct udf_node **node)
{
	*node = (struct udf_node *)(unsigned long)(fi->fh & ULONG_MAX);
	return (0);
}

static int
set_node(struct fuse_file_info *fi, const struct udf_node *node)
{
	fi->fh = (unsigned long)node;
	return (0);
}

static int
decompose_path(const char *path, char **ppath, char **cpath)
{
	char *p;

	*ppath = strdup(path);
	if (*ppath == NULL)
		return (ENOMEM);

	*cpath = NULL;
	for (p = *ppath; *p; ++p) {
		if (*p == '/')
			*cpath = p + 1;
	}
	if (*cpath != NULL)
		*(*cpath - 1) = 0;
	return (0);
}

static int
udf_lookup(struct udf_node *dir_node, struct udf_node **subnode, const char *name)
{
	struct fileid_desc    *fid;
	struct long_ad         udf_icbptr;
	int lb_size, found, error;

	*subnode = NULL;
	lb_size = dir_node->udf_log_vol->lb_size;

	fid = malloc(lb_size);
	if (fid == NULL)
		return ENOMEM;

	error = udf_lookup_name_in_dir(dir_node, name, strlen(name), &udf_icbptr, fid, &found);
	if (!error) {
		error = ENOENT;
		if (found)
			error = udf_readin_udf_node(dir_node, &udf_icbptr, fid, subnode);
	}

	free(fid);
	return error;
}

static int
lookup_pathname(struct udf_node **node, const char *path)
{
	struct udf_node *subnode;
	int error = 0;
	char *fullpath;  /* copy of given path to work on */
	char *cmp;       /* subcomponent name */
	char *p;

	assert(path);

	*node = root_node;

	/* XXX if string like /aaa is given to decompose_path() above
	 * it will split it into empty ("") parent path and "aaa"
	 * child path.
	 * Thus, treat empty path as path to the root node.
	 */

	if (*path == 0)
		return (0);

	fullpath = strdup(path);
	if (fullpath == NULL)
		return (ENOMEM);

	/* skip leading slash */
	cmp = fullpath + 1;
	p = cmp;

	while (*p != 0) {
		while (*p != '/' && *p != 0)
			++p;

		if (*p == '/') {
			/* if we found slash terminate component name
			 * and jump onto begiing of the next component
			 */
			*p = 0;
			++p;
		}

		error = udf_lookup(*node, &subnode, cmp);
		if (error)
			break;

		cmp = p;
		*node = subnode;
	}

	free(fullpath);
	return (error);
}

static void
fixup_stat(const struct udf_node *node, struct stat *stat)
{
	struct fuse_context *ctx;

	ctx = fuse_get_context();

	/* XXX if this is root node, set uid/gid to user that mounted this fs
	 * also reset root dir mode from 0777 to 0755 (XXX)
	 * how to get original mode of the mount-point ?
	 */

	if (node == root_node) {
		stat->st_uid = ctx->uid;
		stat->st_gid = ctx->gid;
		stat->st_mode |= S_IRWXU | S_IRWXG | S_IRWXO;
		stat->st_mode &= ~(S_IWGRP | S_IWOTH);
	}
	else {
		/* reset uid/gid of -1 to user that mounted this fs */
		/* -1 is set on OSes that do not have uid/gid concept */
		if (stat->st_uid == -1)
			stat->st_uid = ctx->uid;
		if (stat->st_gid == -1)
			stat->st_gid = ctx->gid;
	}

	/* XXX inum set by udfclient causes kernel panic,
	 * maybe it is because of inum of zero set for root node ?
	 */
	stat->st_ino = node->hashkey;
}

static int
udfclientfs_getattr(const char *path, struct stat *stat)
{
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) == 0) {
		error = udf_getattr(node, stat);
		fixup_stat(node, stat);
	}
	return (-error);
}

static int
udfclientfs_chmod(const char *path, mode_t mode)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	} 
	else {
		/* change only access mode, not file type! */
		mode &= ~S_IFMT;
		stat.st_mode &= S_IFMT;
		stat.st_mode |= mode;
		error = udf_setattr(node, &stat);
	}
	return (-error);
}

static int
udfclientfs_chown(const char *path, uid_t uid, gid_t gid)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	} 
	else {
		stat.st_uid = uid;
		stat.st_gid = gid;
		error = udf_setattr(node, &stat);
	}
	return (-error);
}

static int
udfclientfs_utime(const char *path, struct utimbuf *times)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	} 
	else {
		stat.st_atime = times->actime;
		stat.st_mtime = times->modtime;
		error = udf_setattr(node, &stat);
	}
	return (-error);
}

static int
udfclientfs_truncate(const char *path, off_t length)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	} 
	else if(length > stat.st_size || length % stat.st_blksize != 0) {
		error = EINVAL;
	}
	else
		error = udf_truncate_node(node, length);

	return (-error);
}

static int
udfclientfs_mknod(const char *path, mode_t mode, dev_t unused)
{
	struct stat stat;
	struct udf_node *pnode;
	struct udf_node *node;
	struct fuse_context *ctx;
	int error;
	char *ppath;
	char *cpath;

	if(udfclientfs_debug)
		printf("mknod: mode = 0%06o\n", mode);

	if ((mode & 0170000) == 0)
		mode |= 0100000;
	else if (S_ISREG(mode) == 0)
		return (-EINVAL);

	ctx = fuse_get_context();

	memset(&stat, 0, sizeof(stat));

	stat.st_mode = mode;
	stat.st_uid = ctx->uid;
	stat.st_gid = ctx->gid;

	if ((error = decompose_path(path, &ppath, &cpath)) != 0) {
	}
	else if	((error = lookup_pathname(&pnode, ppath)) != 0) {
	}
	else
		error = udf_create_file(pnode, cpath, &stat, &node);

	free(ppath);
	return (-error);
}

static int
udfclientfs_mkdir(const char *path, mode_t mode)
{
	struct stat stat;
	struct udf_node *pnode;
	struct udf_node *node;
	struct fuse_context *ctx;
	int error;
	char *ppath;
	char *cpath;

	if(S_ISDIR(mode) == 0)
		return (-EINVAL);

	ctx = fuse_get_context();

	memset(&stat, 0, sizeof(stat));
	stat.st_mode = mode;
	stat.st_uid = ctx->uid;
	stat.st_gid = ctx->gid;

	if ((error = decompose_path(path, &ppath, &cpath)) != 0) {
	}
	else if	((error = lookup_pathname(&pnode, ppath)) != 0) {
	}
	else
		error = udf_create_directory(pnode, cpath, &stat, &node);

	free(ppath);
	return (-error);
}

static int
udfclientfs_unlink(const char *path)
{
	struct udf_node *pnode;
	struct udf_node *node;
	int error;
	char *ppath;
	char *cpath;

	if ((error = decompose_path(path, &ppath, &cpath)) != 0) {
	}
	else if	((error = lookup_pathname(&pnode, ppath)) != 0) {
	}
	else if	((error = lookup_pathname(&node, path)) != 0) {
	}
	else
		error = udf_remove_file(pnode, node, cpath);

	free(ppath);
	return (-error);
}

static int
udfclientfs_rmdir(const char *path)
{
	struct udf_node *pnode;
	struct udf_node *node;
	int error;
	char *ppath;
	char *cpath;

	if ((error = decompose_path(path, &ppath, &cpath)) != 0) {
	}
	else if	((error = lookup_pathname(&pnode, ppath)) != 0) {
	}
	else if	((error = lookup_pathname(&node, path)) != 0) {
	}
	else
		error = udf_remove_directory(pnode, node, cpath);

	free(ppath);
	return (-error);
}

static int
udfclientfs_rename(const char *from, const char *to)
{
	struct udf_node *from_pnode;
	struct udf_node *to_pnode;
	struct udf_node *from_node;
	struct udf_node *to_node;
	int error;
	char *from_ppath;
	char *from_cpath;
	char *to_ppath;
	char *to_cpath;

	if ((error = decompose_path(from, &from_ppath, &from_cpath)) != 0) {
	}
	else if ((error = decompose_path(to, &to_ppath, &to_cpath)) != 0) {
	}
	else if ((error = lookup_pathname(&from_pnode, from_ppath)) != 0) {
	}
	else if ((error = lookup_pathname(&from_node, from)) != 0) {
	}
	else if ((error = lookup_pathname(&to_pnode, to_ppath)) != 0) {
	}
	else {
		/* target need not exist */
		error = lookup_pathname(&to_node, to);
		if(error != 0)
			to_node = NULL;
		error = udf_rename(from_pnode, from_node, from_cpath, to_pnode, to_node, to_cpath);
	}

	free(from_ppath);
	free(to_ppath);
	return (-error);
}

static int
udfclientfs_open(const char *path, struct fuse_file_info *fi)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	}
	else if ((error = set_node(fi, node)) != 0) {
	}
	else
		node->hold++;

	return (-error);
}

static int
udfclientfs_release(const char *path, struct fuse_file_info *fi)
{
	struct udf_node *node;
	int error;

	if ((error = get_node(fi, &node)) == 0)
		node->hold--;
	return (-error);
}

static int
udfclientfs_read(const char *path, char *rbuf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct uio file_uio;
	struct iovec file_iov;
	struct udf_node *node;
	off_t fsize;
	off_t remainder;
	int error;

	if((error = get_node(fi, &node)) != 0)
		return (-error);

	fsize = node->stat.st_size;
	remainder = fsize - offset;

	if (remainder == 0 || size == 0)
		return (0);
	if (remainder < 0)
		return (-EIO);
	if (remainder < size)
		size = remainder;

	bzero(&file_uio, sizeof(struct uio));

	file_uio.uio_rw = UIO_WRITE;
	file_uio.uio_iovcnt = 1;
	file_uio.uio_iov = &file_iov;

	file_uio.uio_offset = offset;
	file_uio.uio_resid = size;
	file_uio.uio_iov->iov_base = rbuf;
	file_uio.uio_iov->iov_len = size;

	error = udf_read_file_part_uio(node, (char*)path, UDF_C_USERDATA, &file_uio);
	if (error == 0)
		return (size);
	return (-error);
}

static int
udfclientfs_write(const char *path, const char *wbuf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct uio file_uio;
	struct iovec file_iov;
	struct udf_node *node;
	int error;

	if (size == 0)
		return (0);

	if((error = get_node(fi, &node)) != 0)
		return (-error);

	bzero(&file_uio, sizeof(struct uio));

	file_uio.uio_rw = UIO_READ;
	file_uio.uio_iovcnt = 1;
	file_uio.uio_iov = &file_iov;

	file_uio.uio_offset = offset;
	file_uio.uio_resid = size;
	file_uio.uio_iov->iov_base = (char*)wbuf;
	file_uio.uio_iov->iov_len = size;

	error = udf_write_file_part_uio(node, (char*)path, UDF_C_USERDATA, &file_uio);
	if (error == 0)
		return (size);
	return (-error);
}

static int
udfclientfs_opendir(const char *path, struct fuse_file_info *fi)
{
	struct stat stat;
	struct udf_node *node;
	int error;

	if ((error = lookup_pathname(&node, path)) != 0) {
	}
	else if ((error = udf_getattr(node, &stat)) != 0) {
	}
	else if ((error = set_node(fi, node)) != 0) {
	}
	else
		node->hold++;

	return (-error);
}

static int
udfclientfs_releasedir(const char *path, struct fuse_file_info *fi)
{
	struct udf_node *node;
	int error;

	if ((error = get_node(fi, &node)) == 0)
		node->hold--;
	return (-error);
}

static int
udfclientfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi)
{
	struct uio dir_uio;
	struct iovec dir_iov;
	struct stat stat;
	struct dirent *dirent;
	struct udf_node *node;
	struct udf_node *entry_node;
	static const int DIR_BUFFER_SIZE = 128 * sizeof(struct dirent);
	char dir_buf[DIR_BUFFER_SIZE];
	off_t off;
	int eof;
	int pos;
	int error;

	if ((error = get_node(fi, &node)) != 0)
		return (-error);

	bzero(&dir_uio, sizeof(struct uio));
	bzero(dir_buf, DIR_BUFFER_SIZE);

	dir_uio.uio_rw = UIO_WRITE;
	dir_uio.uio_iovcnt = 1;
	dir_uio.uio_iov = &dir_iov;

	dir_uio.uio_offset = offset;
	dir_uio.uio_resid = DIR_BUFFER_SIZE;
	dir_uio.uio_iov->iov_base = dir_buf;
	dir_uio.uio_iov->iov_len = DIR_BUFFER_SIZE;

	if ((error = udf_readdir(node, &dir_uio, &eof)) != 0)
		return (-error);

	/* XXX because of internal peculiarity of udfclient uio_offset is not updated
	 * when all entries are read from a directory and eof flag is set to 1.
	 * Thus we must ensure that last call to filler within this function passes
	 * uio_offset as offset, if we have more directory entries left, otherwise
	 * we must pass node->stat.st_size (maximum physical offset).
	 * All preceeding calls to filler may have any (resonable) arbitrary values,
	 * for simplicity we make that value the same for all calls - FUSE seems to
	 * not object to this.
	 */
	off = eof ? node->stat.st_size : dir_uio.uio_offset;
	for (pos = 0; pos < DIR_BUFFER_SIZE - dir_uio.uio_resid; pos += sizeof(struct dirent)) {
		dirent = (struct dirent *) (dir_buf + pos);
		error = udf_lookup(node, &entry_node, dirent->d_name);
		if (error == 0) {
			error = udf_getattr(entry_node, &stat);
			fixup_stat(entry_node, &stat);
		}
		if (filler(buf, dirent->d_name, error ? NULL : &stat, off) != 0)
			break;
	}

	return (0);
}

/* XXX full fs sync is done here */
static int
udfclientfs_fsync(const char *unused1, int unused2, struct fuse_file_info *unused3)
{
	struct udf_discinfo *udf_disc;

	SLIST_FOREACH(udf_disc, &udf_discs_list, next_disc) {
		udf_sync_disc(udf_disc);
	};
	return (0);
}

/* XXX not sure how to correctly set inode stats
 * any other useful information I can set here ?
 */
static int
udfclientfs_statfs(const char *unsed, struct statvfs *buf)
{
	struct udf_log_vol *udf_log_vol = root_node->udf_log_vol;

	buf->f_namemax = 255; /* XXX */

	buf->f_bsize = udf_log_vol->lb_size;
	buf->f_frsize = buf->f_bsize;

	buf->f_blocks = udf_log_vol->total_space / buf->f_bsize;
	buf->f_bfree = udf_log_vol->free_space / buf->f_bsize;
	buf->f_bavail = (udf_log_vol->free_space - udf_log_vol->await_alloc_space) / buf->f_bsize;

	buf->f_files = 999999999; /* XXX */
	buf->f_ffree = 999999999; /* XXX */

	buf->f_flag = 0; /* XXX */
	buf->f_fsid = 0; /* XXX */

	return (0);
}

static void *
udfclientfs_init(struct fuse_conn_info *conn)
{
	(void)conn;
	udf_start_unix_thread();
	return (NULL);
}

static void
udfclientfs_destroy(void *unused)
{
	(void)unused;
	udf_dismount_disc(disc);
}

static struct fuse_operations udfclientfs_oper = {
	.getattr    = udfclientfs_getattr,
	.mknod      = udfclientfs_mknod,
	.mkdir      = udfclientfs_mkdir,
	.unlink     = udfclientfs_unlink,
	.rmdir      = udfclientfs_rmdir,
	.rename     = udfclientfs_rename,
	.chmod      = udfclientfs_chmod,
	.chown      = udfclientfs_chown,
	.truncate   = udfclientfs_truncate,
	.utime      = udfclientfs_utime,
	.open       = udfclientfs_open,
	.fsync      = udfclientfs_fsync,
	.release    = udfclientfs_release,
	.read       = udfclientfs_read,
	.write      = udfclientfs_write,
	.statfs     = udfclientfs_statfs,
	.opendir    = udfclientfs_opendir,
	.releasedir = udfclientfs_releasedir,
	.readdir    = udfclientfs_readdir,
	.init       = udfclientfs_init,
	.destroy    = udfclientfs_destroy,
};

static void
usage(char *name)
{
#if 0
	const char *fusehelp[] = { name, "-ho", NULL };
#endif
	fprintf(stderr, "Usage %s [options] special node\n"
			"\n"
			"%s-specific options:\n"
			"\t-F | -o force\tforce r/w mount even if filesystem is marked dirty (use with caution)\n"
			"\t-r | -o ro\tread-only mount\n"
			"\t-d\t\tdebug mode\n"
			"\n",
	    name, name);
#if 0
	fuse_main(2, (char **) fusehelp, &udfclientfs_oper, NULL);
#endif
	exit(1);
}

static void
parse_option_list(char* in_list, char** out_list, int *mnt_flags_p)
{
	char *opt = in_list;
	char *p = opt;

	/* reserve space for a new string and fast-clear it,
	 * this size should be enough as we can only drop substrings.
	 */
	*out_list = (char*)malloc(strlen(in_list));
	**out_list = 0;

	while (*p != 0) {
		for (; *p != 0; ++p) {
			if (*p == ',') {
				*p = 0;
				++p;
				break;
			}
		}

		if (strcmp(opt, "force") == 0) {
			/* this is udfclientfs option, not passed to fuse */
			*mnt_flags_p |= UDF_MNT_FORCE;
		}
		else {
			if (strcmp(opt, "ro") == 0)
				*mnt_flags_p |= UDF_MNT_RDONLY;
			/* add comma if we have any accumulated options */
			if (**out_list != 0)
				strcat(*out_list, ",");
			strcat(*out_list, opt);
		}

		opt = p;
	}

	/* return NULL if resulting options list is empty */
	if (**out_list == 0) {
		free(*out_list);
		*out_list = NULL;
	}
}

int main(int argc, char *argv[])
{
	static const int MAX_FUSE_ARGC = 64;

	char *dev;
	char *mnt_point;
	char *progname;
	char *fuseargv[MAX_FUSE_ARGC];
	int fuseargc = 0;
	int mnt_flags = 0;
	int error;
	char ch;

	fuseargv[fuseargc++] = progname = basename(argv[0]);

	/* explicitely parse only options that are "interesting" to us,
	 * pass-through all other options
	 */
	/* TODO: change this stuff to use new FUSE option parsing API */
	while ((ch = getopt(argc, argv, ":dFo:r")) != -1) {
		switch (ch) {
		case 'd':
			udfclientfs_debug = 1;
			fuseargv[fuseargc++] = strdup("-d");
			break;
		case 'F':
			mnt_flags |= UDF_MNT_FORCE;
			break;
		case 'o':
			/* put results of option list parsing where they should belong
			 * but only add "-o" and increment argc counter if that list
			 * is not empty
			 */
			parse_option_list(optarg, &fuseargv[fuseargc + 1], &mnt_flags);
			if (fuseargv[fuseargc + 1] != NULL) {
				fuseargv[fuseargc] = strdup("-o");
				fuseargc += 2;
			}
			break;
		case 'r':
			mnt_flags |= UDF_MNT_RDONLY;
			fuseargv[fuseargc++] = strdup("-r");
			break;
		case '?':
			fuseargv[fuseargc] = (char*)malloc(3);
			fuseargv[fuseargc][0] = '-';
			fuseargv[fuseargc][1] = optopt;
			fuseargv[fuseargc][2] = 0;
			fuseargc++;
			break;
		case ':':
		default:
			fprintf(stderr, "option -%c must have parameters\n\n", optopt);
			usage(progname);
		}

		if(fuseargc >= MAX_FUSE_ARGC - 4) {
			fprintf(stderr, "Too many options\n\n");
			usage(progname);
		}
	}

	argc -= optind;
	argv += optind;

	if(argc != 2)
		usage(progname);

	/* add use_ino fuse option to use our "persistent" inums
	 * instead of dynamically generated by fuse
	 */
	fuseargv[fuseargc++] = strdup("-o");
	fuseargv[fuseargc++] = strdup("use_ino");
	fuseargv[fuseargc++] = strdup("-s");

	dev = argv[0];

	mnt_point = fuseargv[fuseargc++] = argv[1];
	fuseargv[fuseargc] = NULL;

	if ((mnt_flags & UDF_MNT_RDONLY) && (mnt_flags & UDF_MNT_FORCE)) {
		fprintf(stderr, "warning: force option is meaningless with ro mounts\n");
		mnt_flags &= ~UDF_MNT_FORCE;
	}

	error = stat(mnt_point, &mp_stat);
	if (error) {
		fprintf(stderr, "Can't stat mount point %s: %s\n", mnt_point, strerror(error));
		exit(1);
	}

	if (udfclientfs_debug == 0) {
		/* silence udfclient internals */
		int devnull = open(_PATH_DEVNULL, O_WRONLY);
		if (devnull < 0) {
			perror("while opening " _PATH_DEVNULL);
			exit(1);
		}
		dup2(devnull, STDOUT_FILENO);
	}

	printf("Opening device %s\n", dev);
	udf_init();
	error = udf_mount_disc(dev, NULL, 0, mnt_flags, &disc);
	if (error) {
		fprintf(stderr, "Can't open device: %s\n", strerror(error));
		exit(1);
	};

	mp = SLIST_FIRST(&udf_mountables);
	root_node = mp->rootdir_node;

	if(mp == NULL || mp->rootdir_node == NULL) {
		fprintf(stderr, "No UDF filesystem root(s) found on device\n");
	}

	if (mp->writable == 0 && (mnt_flags & UDF_MNT_RDONLY) == 0) {
		fprintf(stderr, "Can't mount device read-write\n");
		exit(1);
	}

	if (mp->writable == 0) {
		disc->recordable = 0;
		disc->rewritable = 0;
	}

	error = fuse_main(fuseargc, fuseargv, &udfclientfs_oper, NULL);
	return (error);
}

